"""
Specialize implementations of ScaleMS operations.

"""
import asyncio
import contextlib
import dataclasses
import inspect
import os
import pathlib
import shutil
import typing

import scalems.subprocess
import scalems.workflow
from scalems.exceptions import InternalError
from scalems.exceptions import ProtocolError
from scalems.subprocess._subprocess import OutputFile
from scalems.unique import next_monotonic_integer as _next_int


@dataclasses.dataclass
class SubprocessInput:
    """Implementation-specific form of input for scalems.subprocess.Subprocess."""

    program: pathlib.Path
    args: typing.Sequence[str]
    stdin: typing.Union[None, typing.TextIO]
    stdout: typing.Union[None, typing.TextIO]
    stderr: typing.Union[None, typing.TextIO]
    env: typing.Union[None, typing.Mapping[str, str]]


class _ExecutionContext:
    """Represent the run time environment for a managed workflow item.

    Provide local definition of evolving concept from local/__init__.py
    """

    workflow_manager: scalems.workflow.WorkflowManager
    identifier: bytes


async def subprocessCoroutine(context: _ExecutionContext, signature: SubprocessInput):
    """Implement Subprocess in the local execution context.

    Use this coroutine as the basis for a Task that will provide the Future[SubprocessResult] interface.
    """
    task_directory = os.path.abspath(context.identifier.hex())

    kwargs = {"stdin": signature.stdin, "stdout": signature.stdout, "stderr": signature.stderr, "env": signature.env}
    if signature.env is not None:
        kwargs["env"] = {key: value for key, value in signature.env.items()}

    # TODO: Institute some checks that the implementation is not using more resources
    #  than it was allocated.
    # TODO: Institute checks that the implementation is not generating workflow dependencies
    #  for itself that cannot be scheduled. (Consider a primitive allowing a task to surrender
    #  its resources and suspend itself, to be reconsidered for scheduling eligibility after
    #  its new dependencies have been evaluated.)
    process = await asyncio.create_subprocess_exec(
        os.fsencode(signature.program),
        *signature.args,
        stdin=signature.stdin,
        stdout=signature.stdout,
        stderr=signature.stderr,
        env=signature.env,
        cwd=task_directory,
    )

    try:
        returncode = await process.wait()
        # TODO: Error handling.
        # For now, assume non-zero exit codes are equivalent to task failure.
    except Exception as e:
        # TODO: Repackage as a plausibly temporary failure?
        raise e
    else:
        # Alternative: Mark task as "done" if no exception was raised.
        if returncode == 0:
            # TODO: How to do the output file staging?
            stdout = pathlib.Path(os.path.abspath(signature.stdout.name))
            stderr = pathlib.Path(os.path.abspath(signature.stderr.name))
            # What do we return for a failed task? Maybe raise?
            result = scalems.subprocess.SubprocessResult(exitcode=returncode, stdout=stdout, stderr=stderr, file={})
            return result
        else:
            raise InternalError("No handler for failed tasks.")


@contextlib.asynccontextmanager
async def input_resource_scope(  # noqa: C901
    context: _ExecutionContext,
    task_input: typing.Union[scalems.subprocess.SubprocessInput, typing.Awaitable[scalems.subprocess.SubprocessInput]],
):
    """Manage the actual execution context of the asyncio.subprocess.Process.

    Translate a scalems.subprocess.SubprocessInput to a local SubprocessInput instance.

    InputResource factory for *subprocess* based implementations.

    TODO: How should this be composed in terms of the context and (local) resource type?
        Note that the (currently unused) *context* parameter is available for dispatching.
    """
    # Await the inputs.
    if inspect.isawaitable(task_input):
        task_input = await task_input
    # TODO: Use an awaitable to deliver the SubprocessInput providing the arguments.
    # Consider providing generic awaitable behavior through a decorator or base class
    # so we don't need to do extra checks and can just await all scalems objects.
    # TODO: What sort of validation or normalization do we want to do for the executable name?
    if not isinstance(task_input, scalems.subprocess.SubprocessInput):
        raise InternalError("Unexpected input type.")
    path = shutil.which(task_input.argv[0])
    if path is None or not os.path.exists(path):
        raise InternalError("Could not find executable. Input should be vetted before this point.")
    program = pathlib.Path(shutil.which(task_input.argv[0]))
    args = list(str(arg) for arg in task_input.argv[1:])

    task_directory = os.path.abspath(context.identifier.hex())
    if os.path.exists(task_directory):
        # check if done
        done_file = os.path.join(task_directory, "done")
        if os.path.exists(done_file):
            raise InternalError("Refusing to prepare input resource scope for a completed task.")
        # check if clean
        # currently deferred to specific files below.
        ...
    else:
        # make task directory
        os.mkdir(task_directory)

    # TODO: Make sure we have a good way to handle inputs from other tasks.
    if task_input.inputs is not None and len(task_input.inputs) > 0:
        for key, value in task_input.inputs.items():
            try:
                if not os.path.exists(value):
                    raise ValueError("File not found: {}".format(value))
            except Exception as e:
                raise TypeError("Invalid input (expected file path): {}".format(repr(value))) from e
            args.extend([key, value])

    if task_input.outputs is not None and len(task_input.inputs) > 0:
        for i, (key, value) in enumerate(task_input.outputs.items()):
            if isinstance(value, OutputFile):
                if value.label is not None:
                    # TODO: Add an item to the graph that depends on the task.
                    ...
                # Generate an appropriate output file name.
                # Use the fingerprint of the task_input, the current operation type identifier,
                # and the sequence number of this output file in the current command.

                # Note that this means we need an additional interaction with the workflow manager
                # in order to have enough context to associate the output with the operation that is
                # consuming these input resources. Alternatively, we could broaden the scope of this
                # context manager to represent all run-time resources, and not just the input resources.

                # input_hash = hash(task_input)
                # input_hash += hash('Subprocess')
                # input_hash += hash(i)
                # basename = hashlib.sha256(bytes(input_hash)).digest().hex()
                # TODO: Handle working directory.
                filename = _next_int().to_bytes(32, "big").hex() + value.suffix

                path = os.path.join(task_directory, filename)
                if os.path.exists(path):
                    raise ProtocolError("Output file already exists but reexecution has not been considered.")
                # There is obviously a race condition / security hole between the time
                # that we check an output file name and execute the command. This is
                # unavoidable with process that expects to create a file with a user-provided name.
            else:
                try:
                    path = pathlib.Path(str(value))
                    if not path.is_absolute():
                        path = os.path.join(task_directory, path)
                    if os.path.exists(path):
                        raise ValueError("Output file already exists: {}".format(path))
                except Exception as e:
                    raise TypeError("Invalid input (expected file path): {}".format(repr(value))) from e
            args.extend([key, path])

    # Warning: If subprocess.Popen receives *env* argument that is not None, it **replaces** the
    # default environment (a duplicate of the caller's environment). We might want to provide
    # fancier semantics to copy or reject select variables from the environment or to
    # dynamically read the default environment at execution time (note that the results of
    # such an operation would not represent a unique result!)
    @contextlib.contextmanager
    def get_env() -> dict:
        # TODO: Use provided environment details.
        yield None

    # Get stdin context manager.
    if task_input.stdin is None:

        @contextlib.contextmanager
        def get_stdin():
            yield None

    else:
        # TODO: Strengthen the typing for stdin parameter.
        if not isinstance(task_input.stdin, (os.PathLike, str, list, tuple)):
            # Note: this is an internal error indicating a bug in SCALEMS.
            raise InternalError("No handler for stdin argument of this form." + repr(task_input.stdin))
        if isinstance(task_input.stdin, (list, tuple)):
            # TODO: Assign appropriate responsibility for maintaining filesystem artifacts.
            infile = os.path.join(task_directory, "stdin")
            with open(infile, "w") as fp:
                # Normalize line endings for local environment.
                fp.writelines([line.rstrip() + "\n" for line in task_input.stdin])
        else:
            infile = os.path.abspath(task_input.stdin)

        def get_stdin():
            return open(os.path.abspath(infile), "r")

    def get_output_file_contextmanager(filename):
        if not isinstance(filename, (os.PathLike, str)):
            # Note: this is an internal error indicating a bug in SCALEMS.
            raise InternalError("No handler for path of this form. " + repr(filename))
        outfile = pathlib.Path(filename)
        if not outfile.is_absolute():
            outfile = os.path.join(task_directory, outfile)
        if os.path.exists(outfile):
            raise InternalError("Dirty working directory is not recoverable in current implementation.")

        def get_openfile():
            return open(outfile, "w")

        return get_openfile

    # Get stdout context manager.
    get_stdout = get_output_file_contextmanager(task_input.stdout or "stdout")

    # Get stderr context manager.
    get_stderr = get_output_file_contextmanager(task_input.stderr or "stderr")

    # Create scoped resources. This depends somewhat on the input.
    # For each non-None stdio stream, we need to provide an open file-like handle.
    # Note: there may be a use case for scoped run-time determination of *env*,
    # but we have not yet allowed for that.
    with get_stdin() as fh_in, get_stdout() as fh_out, get_stderr() as fh_err, get_env() as env:
        try:
            # TODO: Create SubprocessInput with a coroutine so that we can yield an awaitable.
            subprocess_input = SubprocessInput(program, args, stdin=fh_in, stdout=fh_out, stderr=fh_err, env=env)
            # Provide resources to the implementation coroutine.
            yield subprocess_input
        finally:
            # Clean up scoped resources.
            # Deliver output files?
            ...



"""Test the scalems machinery for interacting with the RP raptor facets.

For remote dispatching through RP and Raptor, it can be especially hard to get
good test coverage with any granularity. Here, we do our best to unit test the
important pieces for raptor interaction locally. Full raptor sessions are tested
through :file:`test_rp_exec.py`.
"""

import dataclasses
import json
import logging
import os

import packaging.version
import pytest

import scalems
from scalems.radical.raptor import ClientWorkerRequirements
from scalems.radical.raptor import MasterTaskConfiguration
from scalems.radical.raptor import ScaleMSMaster
from scalems.radical.raptor import ScaleMSWorker
from scalems.radical.raptor import WorkerDescription

try:
    import radical.pilot as rp
except ImportError:
    rp = None
else:
    from scalems.radical.raptor import object_encoder

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

pytestmark = pytest.mark.skipif(condition=rp is None, reason="These tests require RADICAL Pilot.")

client_scalems_version = packaging.version.Version(scalems.__version__)
if client_scalems_version.is_prerelease:
    minimum_scalems_version = client_scalems_version.public
else:
    minimum_scalems_version = client_scalems_version.base_version


@pytest.mark.experimental
def test_master_configuration_details(rp_venv):
    """Test the details needed to launch the master script.

    WARNING: This test is incomplete. We can't actually create a raptor.Master easily,
    but we can check its bits and pieces. This test mostly checks function signatures
    and data structures. It does not produce a functioning Raptor configuration, or
    even an actual RP Session!
    """

    num_workers = 1
    worker_processes = 1
    gpus_per_process = 0

    # TODO: Add additional dependencies that we can infer from the workflow.
    versioned_modules = (("scalems", minimum_scalems_version), ("radical.pilot", rp.version_short))

    # Note that the Worker launch has unspecified results if the `named_env`
    # does not exist and is not scheduled to be created with `prepare_env`.
    # However, we are not currently using `named_env`. See #90.
    configuration = MasterTaskConfiguration(
        worker=ClientWorkerRequirements(
            named_env="scalems_test_ve", cpu_processes=worker_processes, gpus_per_process=gpus_per_process
        ),
        versioned_modules=list(versioned_modules),
    )
    # Note: *named_env* is unused, pending work on #90 and others.

    conf_dict: scalems.radical.raptor._MasterTaskConfigurationDict = dataclasses.asdict(configuration)
    configuration = MasterTaskConfiguration.from_dict(conf_dict)
    assert configuration.versioned_modules == list(versioned_modules)
    assert configuration.worker.named_env == "scalems_test_ve"

    encoded = json.dumps(configuration, default=object_encoder, indent=2)
    configuration = scalems.radical.raptor.MasterTaskConfiguration.from_dict(json.loads(encoded))
    assert configuration.versioned_modules == [list(module_spec) for module_spec in versioned_modules]
    assert configuration.worker.named_env == "scalems_test_ve"

    with pytest.warns(match="raptor.Master base class"):
        master = ScaleMSMaster(configuration)
    with master.configure_worker(configuration.worker) as worker_configs:
        assert len(worker_configs) == num_workers
        for descr in worker_configs:
            descr: WorkerDescription
            assert descr.ranks == worker_processes
            assert descr.raptor_class == ScaleMSWorker.__name__
            assert os.path.exists(descr.raptor_file)
    assert not os.path.exists(descr.raptor_file)



"""Tracking filesystem objects across workflow environments."""
import asyncio
import contextlib
import hashlib
import locale
import mmap
import os
import pathlib
import sys
import tempfile
import urllib.parse

import pytest
import scalems.context
import scalems.exceptions
import scalems.file
import scalems.store
from scalems.identifiers import ResourceIdentifier

sample_text = ("Hi there!", "Hello, World.")

# Get 8 bytes of sample binary data.
sample_value = int(42)
sample_data = sample_value.to_bytes(length=8, byteorder=sys.byteorder)
assert int.from_bytes(sample_data, byteorder=sys.byteorder) == sample_value


@contextlib.contextmanager
def text_file(newline=None):
    # Create sample input file.
    # Get a temp file name.
    with tempfile.NamedTemporaryFile() as f:
        path = pathlib.Path(f.name)
    assert path.is_absolute()
    assert not path.exists()

    try:
        with path.open(mode="w", newline=newline) as fh:
            for line in sample_text:
                fh.write(line)
                fh.write("\n")
        # Check that the file contents are what we expect.
        with open(path, "r") as fh:
            assert all([sample == read.rstrip() for sample, read in zip(sample_text, fh)])
        yield path
    finally:
        # Make sure that the FileStore is now the only source of the file.
        path.unlink()


@contextlib.contextmanager
def binary_file():
    # Create sample input file.
    fd, filename = tempfile.mkstemp(text=True)
    try:
        with os.fdopen(fd, "wb") as fh:
            fh.write(sample_data)
        # Check that the file contents are what we expect.
        with open(filename, "rb") as fh:
            data: bytes = fh.read()
            assert len(data) == len(sample_data)
            assert data == sample_data
            assert int.from_bytes(data, byteorder=sys.byteorder) == sample_value
            yield filename
    finally:
        # Make sure that the FileStore is now the only source of the file.
        os.unlink(filename)


def test_hash():
    """Check our checksum strategy for logic errors."""
    data1 = sample_value.to_bytes(length=1, byteorder=sys.byteorder)
    data8 = sample_data
    data7 = int.from_bytes(sample_data, byteorder=sys.byteorder).to_bytes(
        length=len(sample_data) - 1, byteorder=sys.byteorder
    )
    assert int.from_bytes(data1, byteorder=sys.byteorder) == sample_value
    assert int.from_bytes(data7, byteorder=sys.byteorder) == sample_value
    assert int.from_bytes(data8, byteorder=sys.byteorder) == sample_value
    assert data1 != data7
    assert data7 != data8

    null_digest = hashlib.sha256().digest()
    digest7 = hashlib.sha256(data7).digest()
    digest8 = hashlib.sha256(data8).digest()
    assert digest8 != null_digest
    assert digest7 != null_digest
    assert digest7 != digest8

    a = int(0).to_bytes(length=8, byteorder=sys.byteorder)
    b = (int.from_bytes(a, byteorder=sys.byteorder) + 1).to_bytes(length=8, byteorder=sys.byteorder)
    assert a != b

    # Check some assumptions about our hashing of binary and text files.
    native_line_ending = os.linesep
    if native_line_ending == "\n":
        nonnative_line_ending = "\r\n"
    else:
        nonnative_line_ending = "\n"
    with text_file(newline=native_line_ending) as filename:
        text_hash_1 = hashlib.sha256()
        with open(filename, "r") as f:
            for line in f:
                text_hash_1.update(line.encode(locale.getpreferredencoding(False)))
    with text_file(newline=nonnative_line_ending) as filename:
        text_hash_2 = hashlib.sha256()
        with open(filename, "r") as f:
            for line in f:
                text_hash_2.update(line.encode(locale.getpreferredencoding(False)))
    assert text_hash_1.digest() == text_hash_2.digest()
    with text_file(newline=nonnative_line_ending) as filename:
        text_hash_3 = hashlib.sha256()
        with open(filename, "r", newline=native_line_ending) as f:
            for line in f:
                text_hash_3.update(line.encode(locale.getpreferredencoding(False)))
    assert text_hash_1.digest() != text_hash_3.digest()

    with text_file(newline=native_line_ending) as filename:
        binary_hash = hashlib.sha256()
        with open(filename, "rb") as f:
            with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as data:
                binary_hash.update(data)
    assert binary_hash.digest() == text_hash_1.digest()
    assert binary_hash.digest() == text_hash_2.digest()
    assert binary_hash.digest() != text_hash_3.digest()


@pytest.mark.asyncio
async def test_simple_text_file(tmp_path):
    manager = scalems.store.FileStoreManager(directory=tmp_path)
    datastore = manager.filestore()
    with text_file() as filename:
        # Add the file to the file store.
        future = asyncio.ensure_future(datastore.add_file(scalems.file.describe_file(filename, mode="r")))
        await future
        assert future.done()
        fileref: scalems.file.AbstractFileReference = future.result()

        with pytest.raises(scalems.exceptions.DuplicateKeyError):
            await datastore.add_file(scalems.file.describe_file(filename, mode="r"))
        with pytest.raises(scalems.exceptions.DuplicateKeyError):
            await datastore.add_file(scalems.file.describe_file(filename, mode="r"), _name="spam")

    assert not os.path.exists(filename)
    assert pathlib.Path(fileref).exists()

    assert fileref.filestore() is datastore
    assert fileref.key() in datastore.files
    assert fileref.is_local()
    assert await fileref.localize() is fileref
    uri = fileref.as_uri()
    path = urllib.parse.urlparse(uri).path
    assert path == str(pathlib.Path(fileref))

    with open(fileref, "r") as fh:
        assert all([sample == read.rstrip() for sample, read in zip(sample_text, fh)])

    key = fileref.key()
    assert isinstance(key, ResourceIdentifier)
    assert len(key.bytes()) > 0

    # Make sure a small change is caught.
    duplicate = pathlib.Path(fileref).name
    try:
        with open(filename, "w") as fh:
            fh.write("\n".join(sample_text))
        with open(filename, "r") as fh1:
            with text_file() as f:
                with open(f, "r") as fh2:
                    assert not all([line1.encode() == line2.encode() for line1, line2 in zip(fh1, fh2)])
        assert key in datastore.files
        assert pathlib.Path(fileref) in datastore.files.values()
        with pytest.raises(scalems.exceptions.DuplicateKeyError):
            await datastore.add_file(scalems.file.describe_file(filename, mode="r"), _name=duplicate)

        fileref: scalems.file.AbstractFileReference = await datastore.add_file(
            scalems.file.describe_file(filename, mode="r")
        )
        assert fileref.key() != key
        assert pathlib.Path(fileref).name != duplicate
        assert os.fspath(fileref) != path

    finally:
        os.unlink(filename)

    manager.close()


@pytest.mark.asyncio
async def test_simple_binary_file(tmp_path):
    manager = scalems.store.FileStoreManager(directory=tmp_path)
    datastore = manager.filestore()
    with binary_file() as filename:
        # Add the file to the file store.
        future = asyncio.ensure_future(datastore.add_file(scalems.file.describe_file(filename, mode="rb")))
        await future
        assert future.done()
        fileref: scalems.file.AbstractFileReference = future.result()

        with pytest.raises(scalems.exceptions.DuplicateKeyError):
            await datastore.add_file(scalems.file.describe_file(filename, mode="rb"))
    assert not os.path.exists(filename)
    assert pathlib.Path(fileref).exists()

    with open(fileref, "rb") as fh:
        data: bytes = fh.read()
        assert int.from_bytes(data, byteorder=sys.byteorder) == sample_value

    key: ResourceIdentifier = fileref.key()
    assert isinstance(key, ResourceIdentifier)
    assert len(str(key)) > 0

    # Make sure a small change is handled properly.
    duplicate = pathlib.Path(fileref).name
    try:
        new_data = int.from_bytes(sample_data, byteorder=sys.byteorder).to_bytes(
            length=len(sample_data) - 1, byteorder=sys.byteorder
        )
        assert int.from_bytes(new_data, byteorder=sys.byteorder) == sample_value
        with open(filename, "wb") as fh:
            fh.write(new_data)
        # We are not allowed to use the same key for a new file.
        with pytest.raises(scalems.exceptions.DuplicateKeyError):
            await datastore.add_file(scalems.file.describe_file(filename, mode="rb"), _name=duplicate)
        # We are allowed to let the filestore generate a new entry for a new file.
        # The path of the file argument does not matter because we rely on
        # fingerprinting the file contents.
        fileref = await datastore.add_file(scalems.file.describe_file(filename, mode="rb"))
        new_key: ResourceIdentifier = fileref.key()
        assert new_key != key
        assert len(bytes(new_key)) > 0
        # We get an exception if we try to add the same file twice through add_file,
        # but we have a utility function to wrap up the logic for that.
        with pytest.raises(scalems.exceptions.DuplicateKeyError):
            await datastore.add_file(scalems.file.describe_file(filename, mode="rb"))
        recycled = await scalems.store.get_file_reference(pathlib.Path(filename), filestore=datastore)
        assert recycled.key() == new_key
    finally:
        os.unlink(filename)
    del manager
    assert datastore.closed


@pytest.mark.asyncio
async def test_dispatching(tmp_path):
    manager = scalems.store.FileStoreManager(directory=tmp_path)
    datastore = manager.filestore()
    with text_file() as textfile:
        fileref = await scalems.store.get_file_reference(textfile, filestore=datastore, mode="r")
    assert fileref.key() in datastore.files
    with pathlib.Path(fileref).open("r") as fh:
        assert all([sample == read.rstrip() for sample, read in zip(sample_text, fh)])
    del manager



"""Exceptions thrown by SCALE-MS are catchable as scalems.ScaleMSException.

Additional common exceptions are defined in this module.
scalems submodules may define additional exceptions, but all will be derived
from exceptions specified in scalems.exceptions.
"""

import logging as _logging
import typing

logger = _logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


class ScaleMSError(Exception):
    """Base exception for scalems package errors.

    Users should be able to use this base class to catch errors
    emitted by SCALE-MS.
    """


class ScaleMSWarning(Warning):
    """Base Warning for scalems package warnings.

    Users and testers should be able to use this base class to filter
    warnings emitted by SCALE-MS.
    """


class InternalError(ScaleMSError):
    """An otherwise unclassifiable error has occurred (a bug).

    Please report the bug at https://github.com/SCALE-MS/scale-ms/issues
    """


class APIError(ScaleMSError):
    """Specified interfaces are being violated."""


class ContextError(ScaleMSError):
    """A Context operation could not be performed."""


class DispatchError(ScaleMSError):
    """SCALE-MS is unable to execute work or manage data in the requested environment."""


class DuplicateKeyError(ScaleMSError):
    """An identifier is being reused in a situation where this is not supported."""


class MissingImplementationError(ScaleMSError):
    """The expected feature is not available.

    This indicates a bug or incomplete implementation. If error message does not
    cite an existing tracked issue, please file a bug report.
    https://github.com/SCALE-MS/scale-ms/issues
    """


class ProtocolError(ScaleMSError):
    """A behavioral protocol has not been followed correctly."""


class ScopeError(ScaleMSError):
    """A command or reference is not valid in the current scope or Context."""


class ProtocolWarning(ScaleMSWarning):
    """Unexpected behavior is detected that is not fatal, but which may indicate a bug."""


def deprecated(explanation: str):
    """Mark a deprecated definition.

    Wraps a callable to issue a DeprecationWarning when called.

    Use as a parameterized decorator::

        @deprecated("func is deprecated because...")
        def func():
            ...

    """
    try:
        _message = str(explanation)
        assert len(_message) > 0
    except Exception as e:
        raise ValueError("`deprecated` decorator needs a *explanation*.") from e

    def decorator(func: typing.Callable):
        import functools

        def deprecation(message):
            import warnings

            warnings.warn(message, DeprecationWarning, stacklevel=2)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            deprecation(_message)
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""Dispatch scalems.exec through RADICAL Pilot.

For testing purposes, the RP dispatching can use a Docker container to
encapsulate the details of the RP-enabled environment, such as the required
MongoDB instance and RADICAL_PILOT_DBURL environment variable.
Refer to the repository directories ``docker`` and ``.github/workflows``.

Note: ``export RADICAL_LOG_LVL=DEBUG`` to enable RP debugging output.
"""

import asyncio
import dataclasses
import json
import os
import typing
from subprocess import CompletedProcess
from subprocess import run as subprocess_run

import packaging.version
import pytest

import scalems
import scalems.call
import scalems.context
import scalems.identifiers
import scalems.messages
import scalems.workflow

if typing.TYPE_CHECKING:
    import radical.utils as ru

try:
    import radical.pilot as rp
except ImportError:
    rp = None
else:
    import scalems.radical
    import scalems.radical.raptor
    import scalems.radical.runtime

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


# TODO: Catch sigint from RP and apply our own timeout.

pytestmark = pytest.mark.skipif(condition=rp is None, reason="These tests require RADICAL Pilot.")

client_scalems_version = packaging.version.Version(scalems.__version__)
if client_scalems_version.is_prerelease:
    minimum_scalems_version = client_scalems_version.public
else:
    minimum_scalems_version = client_scalems_version.base_version


@pytest.mark.asyncio
async def test_raptor_master(pilot_description, rp_venv):
    """Check our ability to launch and interact with a Master task."""
    import radical.pilot as rp

    # Hopefully, this requirement is temporary.
    if rp_venv is None:
        pytest.skip("This test requires a user-provided static RP venv.")

    job_endpoint: ru.Url = rp.utils.misc.get_resource_job_url(
        pilot_description.resource, pilot_description.access_schema
    )
    launch_method = job_endpoint.scheme
    if launch_method == "fork":
        pytest.skip("Raptor is not fully supported with 'fork'-based launch methods.")

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    logging.getLogger("asyncio").setLevel(logging.DEBUG)

    # Configure module.
    params = scalems.radical.runtime.Configuration(
        execution_target=pilot_description.resource,
        target_venv=rp_venv,
        rp_resource_params={"PilotDescription": pilot_description.as_dict()},
        enable_raptor=True,
    )

    timeout = 180
    manager = scalems.radical.workflow_manager(loop)
    with scalems.workflow.scope(manager, close_on_exit=True):
        async with manager.dispatch(params=params) as dispatcher:
            assert isinstance(dispatcher, scalems.radical.runtime.RPDispatchingExecutor)
            logger.debug(f"test_raptor_master Session is {repr(dispatcher.runtime.session)}")

            # Bypass the scalems machinery and submit an instruction directly to the master task.
            # TODO: Use scalems.radical.runtime.submit()
            scheduler: rp.Task = dispatcher.runtime.scheduler
            hello_command_description = rp.TaskDescription(
                from_dict={
                    "scheduler": scheduler.uid,
                    "mode": scalems.radical.raptor.CPI_MESSAGE,
                    "metadata": scalems.messages.HelloCommand().encode(),
                    "uid": f"command-hello-{scalems.identifiers.EphemeralIdentifier()}",
                }
            )
            logger.debug(f"Submitting {str(hello_command_description.as_dict())}")
            (hello_task,) = await asyncio.to_thread(
                dispatcher.runtime.task_manager().submit_tasks, [hello_command_description]
            )
            logger.debug(f"Submitted {str(hello_task.as_dict())}. Waiting...")
            hello_state = await asyncio.to_thread(hello_task.wait, state=rp.FINAL, timeout=timeout)
            logger.debug(str(hello_task.as_dict()))

            stop_command_description = rp.TaskDescription(
                from_dict={
                    "scheduler": scheduler.uid,
                    "mode": scalems.radical.raptor.CPI_MESSAGE,
                    "metadata": scalems.messages.Control.create("stop").encode(),
                    "uid": f"command-stop--{scalems.identifiers.EphemeralIdentifier()}",
                }
            )
            logger.debug(f"Submitting {str(stop_command_description.as_dict())}")
            (stop_task,) = await asyncio.to_thread(
                dispatcher.runtime.task_manager().submit_tasks, [stop_command_description]
            )

            # We expect the status update -> DONE, even if self.stop() was called during result_cb for the task.
            stop_watcher = asyncio.create_task(
                asyncio.to_thread(stop_task.wait, state=rp.FINAL, timeout=timeout), name="stop-watcher"
            )

            scheduler_watcher = asyncio.create_task(
                asyncio.to_thread(scheduler.wait, state=rp.FINAL, timeout=timeout), name="master-watcher"
            )
            # If master task fails, stop-watcher will never complete.
            done, pending = await asyncio.wait(
                (stop_watcher, scheduler_watcher), timeout=timeout, return_when=asyncio.FIRST_COMPLETED
            )

            if scheduler_watcher not in done:
                await asyncio.wait_for(scheduler_watcher, timeout=10)
            logger.debug(f"scheduler-task state: {scheduler.state}")
            if scheduler.state == rp.DONE and stop_watcher in pending:
                # Waiting longer doesn't seem to help.
                # logger.debug("Waiting a little longer for the stop task to wrap up.")
                # await asyncio.wait_for(stop_watcher, timeout=timeout)
                # TODO(#289) Reconcile expectations regarding stop-task state updates.
                # The prescribed behavior is that the task _should_ reach final state.
                # The following assertion will alert us to the bug fix in scalems or rp that
                # is preventing expected behavior. When it ceases to be true, then we should
                # adjust our assumptions about the behavior of tasks that include
                # a `Master.stop()` in the result_cb.
                assert stop_task.state not in rp.FINAL
            if not stop_watcher.done():
                logger.debug(f"Canceling {stop_task}.")
                stop_watcher.cancel()
            logger.debug(f"stop-task state: {stop_task.state}")

            assert scheduler.state == rp.DONE

    assert hello_state == rp.DONE
    assert hello_task.stdout == repr(scalems.radical.raptor.backend_version)
    # Ref https://github.com/SCALE-MS/scale-ms/discussions/268
    assert hello_task.return_value == dataclasses.asdict(scalems.radical.raptor.backend_version)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.asyncio
async def test_worker(pilot_description, rp_venv):
    """Launch the master script and execute a trivial workflow."""

    if rp_venv is None:
        # Be sure to provision the venv.
        pytest.skip("This test requires a user-provided static RP venv.")

    job_endpoint: ru.Url = rp.utils.misc.get_resource_job_url(
        pilot_description.resource, pilot_description.access_schema
    )
    launch_method = job_endpoint.scheme
    if launch_method == "fork":
        pytest.skip("Raptor is not fully supported with 'fork'-based launch methods.")

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    logging.getLogger("asyncio").setLevel(logging.DEBUG)

    params = scalems.radical.runtime.Configuration(
        execution_target=pilot_description.resource,
        target_venv=rp_venv,
        rp_resource_params={"PilotDescription": pilot_description.as_dict()},
        enable_raptor=True,
    )

    # TODO: Make the work representation non-Raptor-specific or decouple
    #  the Raptor oriented representation of Work from the client-side representation.
    work_item = scalems.radical.raptor.ScalemsRaptorWorkItem(
        func=print.__name__, module=print.__module__, args=["hello world"], kwargs={}, comm_arg_name=None
    )

    timeout = 120
    manager = scalems.radical.workflow_manager(loop)
    with scalems.workflow.scope(manager, close_on_exit=True):
        assert not loop.is_closed()
        # Enter the async context manager for the default dispatcher
        async with manager.dispatch(params=params) as dispatcher:
            # We have now been through RPDispatchingExecutor.runtime_startup().
            assert isinstance(dispatcher, scalems.radical.runtime.RPDispatchingExecutor)
            logger.debug(f"Session is {repr(dispatcher.runtime.session)}")

            task_uid = "add_item.scalems-test-worker"

            # Submit a raptor task
            # Bypass the scalems machinery and submit an instruction directly to the master task.
            # TODO: Use scalems.radical.runtime.submit()
            scheduler: rp.Task = dispatcher.runtime.scheduler

            add_item_task_description = rp.TaskDescription()
            add_item_task_description.scheduler = scheduler.uid
            add_item_task_description.uid = task_uid
            add_item_task_description.cpu_processes = 1
            add_item_task_description.cpu_process_type = (rp.SERIAL,)
            add_item_task_description.mode = scalems.radical.raptor.CPI_MESSAGE
            add_item_task_description.metadata = scalems.messages.AddItem(json.dumps(work_item)).encode()

            task_manager = dispatcher.runtime.task_manager()

            submitter = asyncio.create_task(
                asyncio.to_thread(task_manager.submit_tasks, add_item_task_description), name="rp_submit"
            )
            try:
                submitted_rptask: rp.Task = await asyncio.wait_for(submitter, timeout=timeout)
            except asyncio.TimeoutError as e:
                logger.exception(f"Waited more than {timeout} to submit {add_item_task_description}.")
                submitter.cancel()
                raise e

            cpi_command_future: asyncio.Task = await scalems.radical.runtime.rp_task(submitted_rptask)
            try:
                add_item_task: rp.Task = await asyncio.wait_for(cpi_command_future, timeout=timeout)
            except asyncio.TimeoutError as e:
                logger.debug(f"Waited more than {timeout} for {cpi_command_future}: {e}")
                cpi_command_future.cancel("Canceled after waiting too long.")
                raise e
            assert add_item_task.exit_code == 0

            # Ref https://github.com/SCALE-MS/scale-ms/discussions/268
            try:
                rp_release = packaging.version.parse(rp.version_detail).release
            except packaging.version.InvalidVersion:
                # Ref: https://github.com/radical-cybertools/radical.pilot/issues/2807
                rp_release = None
            if rp_release is not None and rp_release[0:2] == (1, 20):
                assert add_item_task.return_value is None
                work_item_task_id = add_item_task.stdout
            else:
                assert add_item_task.return_value is not None
                work_item_task_id = add_item_task.return_value

            logger.debug(f"Master submitted task {work_item_task_id} to Worker.")
            # We now have the ID of the Task supporting the workflow item we added.
            # We can get the result from a report provided by the Master, or we can use
            # some sort of query command (both TBD).
            # TODO(#229): Check an actual data result.
            # TODO: Update the RPDispatchingExecutor to send the stop rpc call.


# def test_rp_raptor_remote_docker(sdist, rp_task_manager):
#     """Test the core RADICAL Pilot functionality that we rely on through ssh-based execution."""
#     import radical.pilot as rp
#     tmgr = rp_task_manager
#
#     # TODO: How can we recover successful workflow stages from previous failed Sessions?
#     #
#     # The client needs to note the sandbox locations from runs. SCALEMS can
#     # then manage / maintain state tracking or state discovery to optimize workflow recovery.
#     # Resumed workflows can make reference to sandboxes from previous sessions
#     # (RCT work in progress: https://github.com/radical-cybertools/radical.pilot/tree/feature/sandboxes
#     # slated for merge in 2021 Q2 to support `sandbox://` URIs).
#
#     # define a raptor.scalems master and launch it within the pilot
#     pwd   = os.path.dirname(__file__)
#     td    = rp.TaskDescription(
#             {
#                 'uid'          :  'raptor.scalems',
#                 'executable'   :  'python3',
#                 'arguments'    : ['./scalems_test_master.py', '%s/scalems_test_cfg.json'  % pwd],
#                 'input_staging': ['%s/scalems_test_cfg.json'  % pwd,
#                                   '%s/scalems_test_master.py' % pwd,
#                                   '%s/scalems_test_worker.py' % pwd]
#             })
#     scheduler = tmgr.submit_tasks(td)
#
#     # define raptor.scalems tasks and submit them to the master
#     tds = list()
#     for i in range(2):
#         uid  = 'scalems.%06d' % i
#         # ------------------------------------------------------------------
#         # work serialization goes here
#         # This dictionary is interpreted by rp.raptor.Master.
#         work = json.dumps({'mode'      :  'call',
#                            'cores'     :  1,
#                            'timeout'   :  10,
#                            'data'      : {'method': 'hello',
#                                           'kwargs': {'world': uid}}})
#         # ------------------------------------------------------------------
#         tds.append(rp.TaskDescription({
#                            'uid'       : uid,
# The *executable* field is ignored by the ScaleMSMaster that receives this submission.
#                            'executable': 'scalems',
#                            'scheduler' : 'raptor.scalems', # 'scheduler' references the task implemented as a
#                            'arguments' : [work]  # Processed by raptor.Master._receive_tasks
#         }))
#
#     tasks = tmgr.submit_tasks(tds)
#     assert len(tasks) == len(tds)
#     # 'arguments' gets wrapped in a Request at the Master by _receive, then
#     # picked up by the Worker in _request_cb. Then picked up in forked interpreter
#     # by Worker._dispatch, which checks the *mode* of the Request and dispatches
#     # according to native or registered mode implementations. (e.g. 'call' (native) or 'scalems')
#
#     # task process is launched with Python multiprocessing (native) module and added to self._pool.
#     # When the task runs, it's result triggers _result_cb
#
#     # wait for *those* tasks to complete and report results
#     tmgr.wait_tasks(uids=[t.uid for t in tasks])
#
#     # Cancel the master.
#     tmgr.cancel_tasks(uids=scheduler.uid)
#     # Cancel blocks until the task is done so the following wait it currently redundant,
#     # but there is a ticket open to change this behavior.
#     # See https://github.com/radical-cybertools/radical.pilot/issues/2336
#     # tmgr.wait_tasks([scheduler.uid])
#
#     for t in tasks:
#         print('%s  %-10s : %s' % (t.uid, t.state, t.stdout))
#         assert t.state == rp.states.DONE
#         assert t.exit_code == 0


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.asyncio
async def test_rp_function(pilot_description, rp_venv, tmp_path):
    """Test our automation for RP Task generation for function calls."""
    # timeout = 180

    # Hopefully, this requirement is temporary.
    if rp_venv is None:
        pytest.skip("This test requires a user-provided static RP venv.")

    job_endpoint: ru.Url = rp.utils.misc.get_resource_job_url(
        pilot_description.resource, pilot_description.access_schema
    )
    launch_method = job_endpoint.scheme
    if launch_method == "fork":
        pytest.skip("Raptor is not fully supported with 'fork'-based launch methods.")

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    logging.getLogger("asyncio").setLevel(logging.DEBUG)

    # Configure module.
    params = scalems.radical.runtime.Configuration(
        execution_target=pilot_description.resource,
        target_venv=rp_venv,
        rp_resource_params={"PilotDescription": pilot_description.as_dict()},
        enable_raptor=True,
    )

    # Test RPDispatcher context
    manager = scalems.radical.workflow_manager(loop)

    with scalems.workflow.scope(manager, close_on_exit=True):
        assert not loop.is_closed()
        # TODO: Further simplification. E.g.
        #     call_ref = manager.submit(function_call(**sample_call_input))
        #     return_value = call_ref.result()

        async with manager.dispatch(params=params) as dispatcher:
            assert isinstance(dispatcher, scalems.radical.runtime.RPDispatchingExecutor)
            logger.debug(f"exec_rp Session is {repr(dispatcher.runtime.session)}")
            task_uid = "test_rp_function-1"

            call_handle: scalems.call._Subprocess = await scalems.call.function_call_to_subprocess(
                func=subprocess_run,
                kwargs={"args": ["/bin/echo", "hello", "world"], "capture_output": True},
                label=task_uid,
                manager=manager,
                requirements=None,
            )
            # Note: the Master (and Worker) have already started, but may not be
            # commensurate with *requirements*.
            # Resource constraints:
            # * the Master uses one of the cores available to the Pilot,
            #   so it is not available to Tasks.
            # * Tasks cannot span Workers, so we need to make sure that we
            #   provision a sufficiently large Worker.
            # * raptor does not support the memory/disk task constraints.
            # * GPUs: gpus-per-rank is not well explored in raptor. deviations unknown.
            # * nodes: Workers may span nodes, so this shouldn't be a problem.
            # Other set-up details:
            # * pre_exec needs to happen on the Worker, not the Task
            #   (default scalems pre_exec is already handled this way.
            #   We can add a check that user has not extended it until we can
            #   update the Worker provisioning.).
            # * For 0th step, we can provision one Worker with all resources.
            # * For immediate follow-up: Worker provisioning needs to be delayed,
            #   and carried out with respect to the work load.

            rp_task_result: scalems.radical.runtime.RPTaskResult = await scalems.radical.runtime.subprocess_to_rp_task(
                call_handle, dispatcher=dispatcher
            )

            call_result: scalems.call.CallResult = await scalems.radical.runtime.wrapped_function_result_from_rp_task(
                call_handle, rp_task_result
            )

    completed_process: CompletedProcess = call_result.return_value
    assert "hello world" in completed_process.stdout.decode(encoding="utf8")


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.asyncio
async def test_rp_executable(pilot_description, rp_venv):
    """Test our scalems.executable implementation for RADICAL Pilot."""
    import radical.pilot as rp

    # Hopefully, this requirement is temporary.
    if rp_venv is None:
        pytest.skip("This test requires a user-provided static RP venv.")

    original_context = scalems.workflow.get_scope()
    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    logging.getLogger("asyncio").setLevel(logging.DEBUG)

    # Configure module.
    params = scalems.radical.runtime.Configuration(
        execution_target=pilot_description.resource,
        target_venv=rp_venv,
        rp_resource_params={"PilotDescription": pilot_description.as_dict()},
    )

    # Test RPDispatcher context
    manager = scalems.radical.workflow_manager(loop)

    with scalems.workflow.scope(manager, close_on_exit=True):
        assert not loop.is_closed()
        # Test a command from before entering the dispatch context
        cmd1 = scalems.executable(("/bin/echo", "hello", "world"), stdout="stdout1.txt")
        # Enter the async context manager for the default dispatcher
        async with manager.dispatch(params=params) as dispatcher:
            assert isinstance(dispatcher, scalems.radical.runtime.RPDispatchingExecutor)
            logger.debug(f"exec_rp Session is {repr(dispatcher.runtime.session)}")
            # Test a command issued after entering the dispatch context.
            # TODO: Check data flow dependency.
            # cmd2 = scalems.executable(("/bin/cat", cmd1.stdout), stdout="stdout2.txt")
            # TODO: Check *stdin* shim.
            # cmd2 = scalems.executable(("/bin/cat", "-"), stdin=cmd1.stdout, stdout="stdout2.txt")
            cmd2 = scalems.executable(("/bin/echo", "hello", "world"), stdout="stdout2.txt")
            # TODO: Clarify whether/how result() method should work in this scope.
            # TODO: Make scalems.wait(cmd) work as expected in this scope.
        assert cmd1.done()
        assert cmd2.done()
        logger.debug(cmd1.result())
        logger.debug(cmd2.result())

    # TODO: Output typing.
    out1: rp.Task = cmd1.result()
    for output in out1.description["output_staging"]:
        assert os.path.exists(output["target"])
    out2: rp.Task = cmd2.result()
    for output in out2.description["output_staging"]:
        assert os.path.exists(output["target"])
        if output["target"].endswith("stdout"):
            with open(output["target"], "r") as fh:
                line = fh.readline()
                assert line.rstrip() == "hello world"

    # Test active context scoping.
    assert scalems.workflow.get_scope() is original_context
    assert not loop.is_closed()


@pytest.mark.skip(reason="Unimplemented.")
@pytest.mark.asyncio
async def test_batch():
    """Run a batch of uncoupled tasks, dispatched through RP."""
    assert False
